@@ -28,10 +28,12 @@
 **
 ****************************************************************************/
 
+var Codesign = require("../codesign/codesign.js");
 var Cpp = require("cpp.js");
 var File = require("qbs.File");
 var FileInfo = require("qbs.FileInfo");
 var ModUtils = require("qbs.ModUtils");
+var TextFile = require("qbs.TextFile");
 var Utilities = require("qbs.Utilities");
 var WindowsUtils = require("qbs.WindowsUtils");
 
@@ -78,12 +80,37 @@ function hasCxx17Option(input)
 {
     // Probably this is not the earliest version to support the flag, but we have tested this one
     // and it's a pain to find out the exact minimum.
-    return Utilities.versionCompare(input.cpp.compilerVersion, "19.12.25831") >= 0
-            || (input.qbs.toolchain.contains("clang-cl") && input.cpp.compilerVersionMajor >= 7);
+    return (input.qbs.toolchain.includes("clang-cl") && input.cpp.compilerVersionMajor >= 7)
+        || Utilities.versionCompare(input.cpp.compilerVersion, "19.12.25831") >= 0;
+}
+
+function hasZCplusPlusOption(input)
+{
+    // /Zc:__cplusplus is supported starting from Visual Studio 15.7
+    // Looks like closest MSVC version is 14.14.26428 (cl ver 19.14.26433)
+    // At least, this version is tested
+    // https://docs.microsoft.com/en-us/cpp/build/reference/zc-cplusplus
+    // clang-cl supports this option starting around-ish versions 8/9, but it
+    // ignores this option, so this doesn't really matter
+    // https://reviews.llvm.org/D45877
+    return (input.qbs.toolchain.includes("clang-cl") && input.cpp.compilerVersionMajor >= 9)
+        || Utilities.versionCompare(input.cpp.compilerVersion, "19.14.26433") >= 0;
+}
+
+function hasCxx20Option(input)
+{
+    return (input.qbs.toolchain.includes("clang-cl") && input.cpp.compilerVersionMajor >= 13)
+        || Utilities.versionCompare(input.cpp.compilerVersion, "19.29.30133.0") >= 0;
+}
+
+function hasCVerOption(input)
+{
+    return (input.qbs.toolchain.includes("clang-cl") && input.cpp.compilerVersionMajor >= 13)
+        || Utilities.versionCompare(input.cpp.compilerVersion, "19.29.30138.0") >= 0;
 }
 
 function supportsExternalIncludesOption(input) {
-    if (input.qbs.toolchain.contains("clang-cl"))
+    if (input.qbs.toolchain.includes("clang-cl"))
         return false; // Exclude clang-cl.
     // This option was introcuded since MSVC 2017 v15.6 (aka _MSC_VER 19.13).
     // But due to some MSVC bugs:
@@ -92,16 +119,15 @@ function supportsExternalIncludesOption(input) {
     return Utilities.versionCompare(input.cpp.compilerVersion, "19.16") >= 0;
 }
 
-function addLanguageVersionFlag(input, args) {
+function addCxxLanguageVersionFlag(input, args) {
     var cxxVersion = Cpp.languageVersion(input.cpp.cxxLanguageVersion,
-                                         ["c++17", "c++14", "c++11", "c++98"], "C++");
+            ["c++23", "c++20", "c++17", "c++14", "c++11", "c++98"], "C++");
     if (!cxxVersion)
         return;
 
-    // Visual C++ 2013, Update 3
-    var hasStdOption = Utilities.versionCompare(input.cpp.compilerVersion, "18.00.30723") >= 0
-            // or clang-cl
-            || input.qbs.toolchain.contains("clang-cl");
+    // Visual C++ 2013, Update 3 or clang-cl
+    var hasStdOption = input.qbs.toolchain.includes("clang-cl")
+        || Utilities.versionCompare(input.cpp.compilerVersion, "18.00.30723") >= 0;
     if (!hasStdOption)
         return;
 
@@ -110,14 +136,35 @@ function addLanguageVersionFlag(input, args) {
         flag = "/std:c++14";
     else if (cxxVersion === "c++17" && hasCxx17Option(input))
         flag = "/std:c++17";
+    else if (cxxVersion === "c++20" && hasCxx20Option(input))
+        flag = "/std:c++20";
     else if (cxxVersion !== "c++11" && cxxVersion !== "c++98")
         flag = "/std:c++latest";
     if (flag)
         args.push(flag);
 }
 
+function addCLanguageVersionFlag(input, args) {
+    var cVersion = Cpp.languageVersion(input.cpp.cLanguageVersion,
+            ["c17", "c11"], "C");
+    if (!cVersion)
+        return;
+
+    var hasStdOption = hasCVerOption(input);
+    if (!hasStdOption)
+        return;
+
+    var flag;
+    if (cVersion === "c17")
+        flag = "/std:c17";
+    else if (cVersion === "c11")
+        flag = "/std:c11";
+    if (flag)
+        args.push(flag);
+}
+
 function handleClangClArchitectureFlags(product, architecture, flags) {
-    if (product.qbs.toolchain.contains("clang-cl")) {
+    if (product.qbs.toolchain.includes("clang-cl")) {
         if (architecture === "x86")
             flags.push("-m32");
         else if (architecture === "x86_64")
@@ -125,7 +172,7 @@ function handleClangClArchitectureFlags(product, architecture, flags) {
     }
 }
 
-function prepareCompiler(project, product, inputs, outputs, input, output, explicitlyDependsOn) {
+function prepareCompilerInternal(project, product, inputs, outputs, input, output, explicitlyDependsOn) {
     var i;
     var debugInformation = input.cpp.debugInformation;
     var args = ['/nologo', '/c']
@@ -134,7 +181,7 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
 
     // Determine which C-language we're compiling
     var tag = ModUtils.fileTagForTargetLanguage(input.fileTags.concat(Object.keys(outputs)));
-    if (!["c", "cpp"].contains(tag))
+    if (!["c", "cpp", "cppm"].includes(tag))
         throw ("unsupported source language");
 
     var enableExceptions = input.cpp.enableExceptions;
@@ -187,9 +234,7 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
         args.push(rtl);
     }
 
-    var driverFlags = product.cpp.driverFlags;
-    if (driverFlags)
-        args = args.concat(driverFlags);
+    args = args.concat(Cpp.collectMiscDriverArguments(product));
 
     // warnings:
     var warningLevel = input.cpp.warningLevel;
@@ -199,37 +244,33 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
         args.push('/Wall')
     if (input.cpp.treatWarningsAsErrors)
         args.push('/WX')
-    var includePaths = input.cpp.includePaths;
-    if (includePaths) {
-        args = args.concat([].uniqueConcat(includePaths).map(function(path) {
-            return '/I' + FileInfo.toWindowsSeparators(path);
-        }));
-    }
 
-    var allSystemIncludePaths = [];
-    var systemIncludePaths = input.cpp.systemIncludePaths;
-    if (systemIncludePaths)
-        allSystemIncludePaths = allSystemIncludePaths.uniqueConcat(systemIncludePaths);
-    var distributionIncludePaths = input.cpp.distributionIncludePaths;
-    if (distributionIncludePaths)
-        allSystemIncludePaths = allSystemIncludePaths.uniqueConcat(distributionIncludePaths);
-    var includeFlag = "/I";
-    if (supportsExternalIncludesOption(input)) {
-        args.push("/experimental:external");
-        includeFlag = "/external:I"
-    }
-    allSystemIncludePaths.forEach(function(path) {
-        args.push(includeFlag, FileInfo.toWindowsSeparators(path)); });
+    var includePaths = Cpp.collectIncludePaths(input);
+    args = args.concat([].uniqueConcat(includePaths).map(function(path) {
+        return input.cpp.includeFlag + FileInfo.toWindowsSeparators(path);
+    }));
 
-    var allDefines = [];
-    var platformDefines = input.cpp.platformDefines;
-    if (platformDefines)
-        allDefines = allDefines.uniqueConcat(platformDefines);
-    var defines = input.cpp.defines;
-    if (defines)
-        allDefines = allDefines.uniqueConcat(defines);
-    for (i in allDefines)
-        args.push('/D' + allDefines[i].replace(/%/g, "%%"));
+    var includeFlag = input.qbs.toolchain.includes("clang-cl")
+        ? input.cpp.systemIncludeFlag : input.cpp.includeFlag;
+    if (!input.qbs.toolchain.includes("clang-cl")) {
+        if (supportsExternalIncludesOption(input)) {
+            args.push("/experimental:external");
+            var enforcesSlashW =
+                    Utilities.versionCompare(input.cpp.compilerVersion, "19.29.30037") >= 0
+            if (enforcesSlashW)
+                args.push("/external:W0")
+            includeFlag = input.cpp.systemIncludeFlag;
+        }
+    }
+    var systemIncludePaths = Cpp.collectSystemIncludePaths(input);
+    args = args.concat([].uniqueConcat(systemIncludePaths).map(function(path) {
+        return includeFlag + FileInfo.toWindowsSeparators(path);
+    }));
+
+    var defines = Cpp.collectDefines(input);
+    args = args.concat([].uniqueConcat(defines).map(function(define) {
+        return input.cpp.defineFlag + define.replace(/%/g, "%%");
+    }));
 
     var minimumWindowsVersion = product.cpp.minimumWindowsVersion;
     if (minimumWindowsVersion) {
@@ -237,7 +278,7 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
         if (hexVersion) {
             var versionDefs = [ 'WINVER', '_WIN32_WINNT', '_WIN32_WINDOWS' ];
             for (i in versionDefs) {
-                args.push('/D' + versionDefs[i] + '=' + hexVersion);
+                args.push(input.cpp.defineFlag + versionDefs[i] + '=' + hexVersion);
             }
         }
     }
@@ -248,21 +289,31 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
     if (input.cpp.generateCompilerListingFiles)
         args.push("/Fa" + FileInfo.toWindowsSeparators(outputs.lst[0].filePath));
 
+    if (input.cpp.enableCxxLanguageMacro && hasZCplusPlusOption(input))
+        args.push("/Zc:__cplusplus");
+
     var objectMap = outputs.obj || outputs.intermediate_obj
     var objOutput = objectMap ? objectMap[0] : undefined
     args.push('/Fo' + FileInfo.toWindowsSeparators(objOutput.filePath))
     args.push(FileInfo.toWindowsSeparators(input.filePath))
 
-    var prefixHeaders = product.cpp.prefixHeaders;
-    for (i in prefixHeaders)
-        args.push("/FI" + FileInfo.toWindowsSeparators(prefixHeaders[i]));
+    var preincludePaths = Cpp.collectPreincludePaths(input);
+    args = args.concat([].uniqueConcat(preincludePaths).map(function(path) {
+        return input.cpp.preincludeFlag + FileInfo.toWindowsSeparators(path);
+    }));
 
     // Language
-    if (tag === "cpp") {
+    if (tag === "cpp" || tag === "cppm") {
         args.push("/TP");
-        addLanguageVersionFlag(input, args);
+        addCxxLanguageVersionFlag(input, args);
     } else if (tag === "c") {
         args.push("/TC");
+        addCLanguageVersionFlag(input, args);
+    }
+
+    var moduleMap = (outputs["modulemap"] || [])[0];
+    if (moduleMap) {
+        args.push("@" + moduleMap.filePath);
     }
 
     // Whether we're compiling a precompiled header or normal source file
@@ -270,7 +321,7 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
     var pchInputs = explicitlyDependsOn[tag + "_pch"];
     if (pchOutput) {
         // create PCH
-        if (input.qbs.toolchain.contains("clang-cl")) {
+        if (input.qbs.toolchain.includes("clang-cl")) {
             // clang-cl does not support /Yc flag without filename
             args.push("/Yc" + FileInfo.toWindowsSeparators(input.filePath));
             // clang-cl complains when pch file is not included
@@ -300,10 +351,7 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
         }
     }
 
-    args = args.concat(ModUtils.moduleProperty(input, 'platformFlags'),
-                       ModUtils.moduleProperty(input, 'flags'),
-                       ModUtils.moduleProperty(input, 'platformFlags', tag),
-                       ModUtils.moduleProperty(input, 'flags', tag));
+    args = args.concat(Cpp.collectMiscCompilerArguments(input, tag));
 
     var compilerPath = product.cpp.compilerPath;
     var wrapperArgs = product.cpp.compilerWrapper;
@@ -329,75 +377,17 @@ function prepareCompiler(project, product, inputs, outputs, input, output, expli
     return [cmd];
 }
 
-function collectLibraryDependencies(product) {
-    var seen = {};
-    var result = [];
-
-    function addFilePath(filePath, wholeArchive, productName) {
-        result.push({ filePath: filePath, wholeArchive: wholeArchive, productName: productName });
-    }
-
-    function addArtifactFilePaths(dep, artifacts) {
-        if (!artifacts)
-            return;
-        var artifactFilePaths = artifacts.map(function(a) { return a.filePath; });
-        var wholeArchive = dep.parameters.cpp && dep.parameters.cpp.linkWholeArchive;
-        var artifactsAreImportLibs = artifacts.length > 0
-                && artifacts[0].fileTags.contains("dynamiclibrary_import");
-        for (var i = 0; i < artifactFilePaths.length; ++i) {
-            addFilePath(artifactFilePaths[i], wholeArchive,
-                        artifactsAreImportLibs ? dep.name : undefined);
-        }
-    }
-
-    function addExternalLibs(obj) {
-        if (!obj.cpp)
-            return;
-        function ensureArray(a) {
-            return Array.isArray(a) ? a : [];
-        }
-        function sanitizedModuleListProperty(obj, moduleName, propertyName) {
-            return ensureArray(ModUtils.sanitizedModuleProperty(obj, moduleName, propertyName));
-        }
-        var externalLibs = [].concat(
-                    sanitizedModuleListProperty(obj, "cpp", "staticLibraries"),
-                    sanitizedModuleListProperty(obj, "cpp", "dynamicLibraries"));
-        externalLibs.forEach(function (libName) {
-            if (!libName.match(/\.lib$/i) && !libName.startsWith('@'))
-                libName += ".lib";
-            addFilePath(libName, false);
-        });
-    }
-
-    function traverse(dep) {
-        if (seen.hasOwnProperty(dep.name))
-            return;
-
-        if (dep.parameters.cpp && dep.parameters.cpp.link === false)
-            return;
-
-        var staticLibraryArtifacts = dep.artifacts["staticlibrary"];
-        var dynamicLibraryArtifacts = staticLibraryArtifacts
-                ? null : dep.artifacts["dynamiclibrary_import"];
-        if (staticLibraryArtifacts) {
-            seen[dep.name] = true;
-            dep.dependencies.forEach(traverse);
-            addArtifactFilePaths(dep, staticLibraryArtifacts);
-            addExternalLibs(dep);
-        } else if (dynamicLibraryArtifacts) {
-            seen[dep.name] = true;
-            addArtifactFilePaths(dep, dynamicLibraryArtifacts);
-        }
-    }
-
-    product.dependencies.forEach(traverse);
-    addExternalLibs(product);
+function prepareCompiler(project, product, inputs, outputs, input, output, explicitlyDependsOn) {
+    var result = Cpp.prepareModules(project, product, inputs, outputs, input, output);
+    result = result.concat(prepareCompilerInternal(
+        project, product, inputs, outputs, input, output, explicitlyDependsOn));
     return result;
 }
 
 function linkerSupportsWholeArchive(product)
 {
-    return Utilities.versionCompare(product.cpp.compilerVersion, "19.0.24215.1") >= 0
+    return product.qbs.toolchainType.includes("clang-cl") ||
+        Utilities.versionCompare(product.cpp.compilerVersion, "19.0.24215.1") >= 0
 }
 
 function handleDiscardProperty(product, flags) {
@@ -413,16 +403,16 @@ function prepareLinker(project, product, inputs, outputs, input, output) {
     var linkDLL = (outputs.dynamiclibrary ? true : false)
     var primaryOutput = (linkDLL ? outputs.dynamiclibrary[0] : outputs.application[0])
     var debugInformation = product.cpp.debugInformation;
-    var additionalManifestInputs = Array.prototype.map.call(inputs["native.pe.manifest"],
+    var additionalManifestInputs = Array.prototype.map.call(inputs["native.pe.manifest"] || [],
         function (a) {
             return a.filePath;
         });
-    var moduleDefinitionInputs = Array.prototype.map.call(inputs["def"],
+    var moduleDefinitionInputs = Array.prototype.map.call(inputs["def"] || [],
         function (a) {
             return a.filePath;
         });
     var generateManifestFiles = !linkDLL && product.cpp.generateManifestFile;
-    var useClangCl = product.qbs.toolchain.contains("clang-cl");
+    var useClangCl = product.qbs.toolchain.includes("clang-cl");
     var canEmbedManifest = useClangCl || product.cpp.compilerVersionMajor >= 17 // VS 2012
 
     var linkerPath = effectiveLinkerPath(product, inputs);
@@ -434,19 +424,15 @@ function prepareLinker(project, product, inputs, outputs, input, output) {
 
     if (useCompilerDriver) {
         args.push('/nologo');
-        var driverFlags = product.cpp.driverFlags;
-        if (driverFlags)
-            args = args.concat(driverFlags);
-        var driverLinkerFlags = product.cpp.driverLinkerFlags;
-        if (driverLinkerFlags)
-            args = args.concat(driverLinkerFlags);
+        args = args.concat(Cpp.collectMiscDriverArguments(product),
+                           Cpp.collectMiscLinkerArguments(product));
     }
 
-    var allInputs = inputs.obj || [];
-    for (i in allInputs) {
-        var fileName = FileInfo.toWindowsSeparators(allInputs[i].filePath)
-        args.push(fileName)
-    }
+    var allInputs = [].concat(Cpp.collectLinkerObjectPaths(inputs),
+                              Cpp.collectResourceObjectPaths(inputs));
+    args = args.concat([].uniqueConcat(allInputs).map(function(path) {
+        return FileInfo.toWindowsSeparators(path);
+    }));
 
     var linkerArgs = ['/nologo']
     if (linkDLL) {
@@ -539,7 +525,7 @@ function prepareLinker(project, product, inputs, outputs, input, output) {
 
     var wholeArchiveSupported = linkerSupportsWholeArchive(product);
     var wholeArchiveRequested = false;
-    var libDeps = collectLibraryDependencies(product);
+    var libDeps = Cpp.collectLibraryDependencies(product);
     var prevLib;
     for (i = 0; i < libDeps.length; ++i) {
         var dep = libDeps[i];
@@ -580,12 +566,12 @@ function prepareLinker(project, product, inputs, outputs, input, output) {
         args.push('/Fe' + linkerOutputNativeFilePath);
     else
         linkerArgs.push('/OUT:' + linkerOutputNativeFilePath);
-    var libraryPaths = product.cpp.libraryPaths;
-    if (libraryPaths)
-        libraryPaths = [].uniqueConcat(libraryPaths);
-    for (i in libraryPaths) {
-        linkerArgs.push('/LIBPATH:' + FileInfo.toWindowsSeparators(libraryPaths[i]))
-    }
+
+    var libraryPaths = Cpp.collectLibraryPaths(product);
+    linkerArgs = linkerArgs.concat([].uniqueConcat(libraryPaths).map(function(path) {
+        return product.cpp.libraryPathFlag + FileInfo.toWindowsSeparators(path);
+    }));
+
     handleDiscardProperty(product, linkerArgs);
     var linkerFlags = product.cpp.platformLinkerFlags.concat(product.cpp.linkerFlags);
     linkerArgs = linkerArgs.concat(linkerFlags);
@@ -594,9 +580,9 @@ function prepareLinker(project, product, inputs, outputs, input, output) {
 
     var wrapperArgs = product.cpp.linkerWrapper;
     if (wrapperArgs && wrapperArgs.length > 0) {
-        linkerArgs.unshift(linkerPath);
+        args.unshift(linkerPath);
         linkerPath = wrapperArgs.shift();
-        linkerArgs = wrapperArgs.concat(linkerArgs);
+        args = wrapperArgs.concat(args);
     }
     var commands = [];
     var warningCmd = new JavaScriptCommand();
@@ -655,6 +641,12 @@ function prepareLinker(project, product, inputs, outputs, input, output) {
         commands.push(cmd);
     }
 
+    if (product.cpp.shouldSignArtifacts) {
+        Array.prototype.push.apply(
+                    commands, Codesign.prepareSigntool(
+                        project, product, inputs, outputs, input, output));
+    }
+
     return commands;
 }
 
@@ -701,3 +693,155 @@ function createRcCommand(binary, input, output, logo) {
 
     return cmd;
 }
+
+function assemblerCommands(project, product, inputs, outputs, input, output, explicitlyDependsOn)
+{
+    var args = ["/nologo", "/c", "/Fo" + FileInfo.toWindowsSeparators(output.filePath)];
+    if (product.cpp.debugInformation)
+        args.push("/Zi");
+    args = args.concat(Cpp.collectMiscAssemblerArguments(input, "asm"));
+    args.push(FileInfo.toWindowsSeparators(input.filePath));
+    var cmd = new Command(product.cpp.assemblerPath, args);
+    cmd.description = "assembling " + input.fileName;
+    cmd.jobPool = "assembler";
+    cmd.inputFileName = input.fileName;
+    cmd.stdoutFilterFunction = function(output) {
+        var lines = output.split("\r\n").filter(function (s) {
+            return !s.endsWith(inputFileName); });
+        return lines.join("\r\n");
+    };
+    return cmd;
+}
+
+function libtoolOutputArtifacts(product)
+{
+    var artifacts = [
+        {
+            fileTags: ["staticlibrary"],
+            filePath: FileInfo.joinPaths(product.destinationDirectory,
+                                         PathTools.staticLibraryFilePath(product))
+        }
+    ];
+    if (product.cpp.debugInformation && product.cpp.separateDebugInformation) {
+        artifacts.push({
+            fileTags: ["debuginfo_cl"],
+            filePath: product.targetName + ".cl" + product.cpp.debugInfoSuffix
+        });
+    }
+    return artifacts;
+}
+
+function libtoolCommands(project, product, inputs, outputs, input, output, explicitlyDependsOn)
+{
+    var args = ['/nologo']
+    var lib = outputs["staticlibrary"][0];
+    var nativeOutputFileName = FileInfo.toWindowsSeparators(lib.filePath)
+    args.push('/OUT:' + nativeOutputFileName)
+    Array.prototype.push.apply(args, product.cpp.archiverFlags);
+    for (var i in inputs.obj) {
+        var fileName = FileInfo.toWindowsSeparators(inputs.obj[i].filePath)
+        args.push(fileName)
+    }
+    for (var i in inputs.res) {
+        var fileName = FileInfo.toWindowsSeparators(inputs.res[i].filePath)
+        args.push(fileName)
+    }
+    var cmd = new Command(product.cpp.archiverPath, args);
+    cmd.description = 'creating ' + lib.fileName;
+    cmd.highlight = 'linker';
+    cmd.jobPool = "linker";
+    cmd.workingDirectory = FileInfo.path(lib.filePath)
+    cmd.responseFileUsagePrefix = '@';
+    return cmd;
+ }
+
+function dllLinkerOutputArtifacts(product)
+{
+    var artifacts = [
+        {
+            fileTags: ["dynamiclibrary"].concat(
+                product.cpp.shouldSignArtifacts ? ["codesign.signed_artifact"] : []),
+            filePath: FileInfo.joinPaths(product.destinationDirectory,
+                                         PathTools.dynamicLibraryFilePath(product))
+        },
+        {
+            fileTags: ["dynamiclibrary_import"],
+            filePath: FileInfo.joinPaths(product.destinationDirectory,
+                                         PathTools.importLibraryFilePath(product)),
+            alwaysUpdated: false
+        }
+    ];
+    if (product.cpp.debugInformation && product.cpp.separateDebugInformation) {
+        var lib = artifacts[0];
+        artifacts.push({
+            fileTags: ["debuginfo_dll"],
+            filePath: lib.filePath.substr(0, lib.filePath.length - 4)
+                      + product.cpp.debugInfoSuffix
+        });
+    }
+    return artifacts;
+}
+
+function appLinkerOutputArtifacts(product)
+{
+    var app = {
+        fileTags: ["application"].concat(
+            product.cpp.shouldSignArtifacts ? ["codesign.signed_artifact"] : []),
+        filePath: FileInfo.joinPaths(
+                      product.destinationDirectory,
+                      PathTools.applicationFilePath(product))
+    };
+    var artifacts = [app];
+    if (product.cpp.debugInformation && product.cpp.separateDebugInformation) {
+        artifacts.push({
+            fileTags: ["debuginfo_app"],
+            filePath: app.filePath.substr(0, app.filePath.length - 4)
+                      + product.cpp.debugInfoSuffix
+        });
+    }
+    if (product.cpp.generateLinkerMapFile) {
+        artifacts.push({
+            fileTags: ["mem_map"],
+            filePath: FileInfo.joinPaths(
+                          product.destinationDirectory,
+                          product.targetName + product.cpp.linkerMapSuffix)
+        });
+    }
+    return artifacts;
+}
+
+function configureStdModules() {
+    try {
+        const modulesJsonPath = FileInfo.joinPaths(_modulesDirPath, "modules.json");
+        if (File.exists(modulesJsonPath)) {
+            const jsonFile = new TextFile(modulesJsonPath, TextFile.ReadOnly);
+            const json = JSON.parse(jsonFile.readAll());
+            jsonFile.close();
+
+            const moduleSources = json["module-sources"];
+            if (moduleSources) {
+                const modules = moduleSources
+                    .filter(function(module) {
+                        if (module === "std.ixx" && (_forceUseImportStd || _forceUseImportStdCompat)) {
+                            return true;
+                        } else if (module === "std.compat.ixx" && _forceUseImportStdCompat) {
+                            return true;
+                        }
+                        return false;
+                    })
+                    .map(function(module) {
+                        return FileInfo.joinPaths(_modulesDirPath, module);
+                    })
+                    .filter(function(module) {
+                        return File.exists(module);
+                    });
+                if (modules.length > 0) {
+                    found = true;
+                    _stdModulesFiles = modules;
+                }
+            }
+        }
+    } catch (e) {
+        console.error("Error reading modules.json: " + e);
+    }
+}
