@@ -92,8 +92,8 @@ static void resolveDepencency(const RawScannedDependency &dependency,
                 dependencyInProduct = foundArtifact;
             } else if (!productOfDependencyIsDependency) {
                 dependencyInOtherProduct = foundArtifact;
-                productOfDependencyIsDependency
-                        = contains(product->dependencies, dependencyInOtherProduct->product.lock());
+                productOfDependencyIsDependency = product->hasDependency(
+                    dependencyInOtherProduct->product.lock());
             }
             break;
         }
@@ -103,9 +103,14 @@ static void resolveDepencency(const RawScannedDependency &dependency,
     }
 
     // prioritize found artifacts
-    if ((result->file = dependencyInProduct)
-        || (result->file = dependencyInOtherProduct)
-        || (result->file = fileDependencyArtifact)) {
+    if (dependencyInProduct)
+        result->file = dependencyInProduct;
+    else if (dependencyInOtherProduct)
+        result->file = dependencyInOtherProduct;
+    else
+        result->file = fileDependencyArtifact;
+
+    if (result->file) {
         result->filePath = result->file->filePath();
 
         if (result->file == dependencyInOtherProduct && !productOfDependencyIsDependency) {
@@ -175,8 +180,8 @@ void InputArtifactScanner::scanForFileDependencies(Artifact *inputArtifact)
         return;
     m_fileTagsForScanner
             = inputArtifact->fileTags().toStringList().join(QLatin1Char(',')).toLatin1();
-    InputArtifactScannerContext::CacheItem *lastPerFileCacheItem = nullptr;
-    InputArtifactScannerContext::CacheItem *lastPerPropsCacheItem = nullptr;
+    InputArtifactScannerContext::ScannerKeyCache *lastPerFileCacheItem = nullptr;
+    InputArtifactScannerContext::ScannerKeyCache *lastPerPropsCacheItem = nullptr;
     while (!filesToScan.empty()) {
         FileResourceBase *fileToBeScanned = filesToScan.takeFirst();
         const QString &filePathToBeScanned = fileToBeScanned->filePath();
@@ -184,7 +189,7 @@ void InputArtifactScanner::scanForFileDependencies(Artifact *inputArtifact)
             continue;
 
         for (DependencyScanner * const scanner : scanners) {
-            InputArtifactScannerContext::CacheItem *cacheItem;
+            InputArtifactScannerContext::ScannerKeyCache *cacheItem;
             if (scanner->cacheIsPerFile()) {
                 if (!lastPerFileCacheItem)
                     lastPerFileCacheItem = &m_context->cachePerFile[inputArtifact];
@@ -196,8 +201,12 @@ void InputArtifactScanner::scanForFileDependencies(Artifact *inputArtifact)
                 }
                 cacheItem = lastPerPropsCacheItem;
             }
-            scanForScannerFileDependencies(scanner, inputArtifact, fileToBeScanned,
-                scanner->recursive() ? &filesToScan : nullptr, (*cacheItem)[scanner->key()]);
+            scanForScannerFileDependencies(
+                scanner,
+                inputArtifact,
+                fileToBeScanned,
+                scanner->recursive() ? &filesToScan : nullptr,
+                (*cacheItem)[scanner->id()]);
         }
     }
 }
@@ -211,61 +220,62 @@ Set<DependencyScanner *> InputArtifactScanner::scannersForArtifact(const Artifac
             = m_context->scannersCache[product];
     for (const FileTag &fileTag : artifact->fileTags()) {
         InputArtifactScannerContext::DependencyScannerCacheItem &cache = scannerCache[fileTag];
-        if (!cache.valid) {
-            cache.valid = true;
+        if (!cache) {
+            QList<DependencyScannerPtr> cacheScanners;
             const auto scanners = ScannerPluginManager::scannersForFileTag(fileTag);
-            transform(scanners, cache.scanners, [](const auto &scanner) {
-                return std::make_shared<PluginDependencyScanner>(scanner); });
+            transform(scanners, cacheScanners, [](const auto &scanner) {
+                return std::make_shared<PluginDependencyScanner>(scanner);
+            });
             for (const ResolvedScannerConstPtr &scanner : product->scanners) {
                 if (scanner->inputs.contains(fileTag)) {
-                    cache.scanners.push_back(
-                                std::make_shared<UserDependencyScanner>(scanner, engine));
+                    cacheScanners.push_back(
+                        std::make_shared<UserDependencyScanner>(scanner, engine));
                     break;
                 }
             }
+            cache = std::move(cacheScanners);
         }
-        for (const DependencyScannerPtr &scanner : std::as_const(cache.scanners))
+        for (const DependencyScannerPtr &scanner : std::as_const(*cache))
             scanners += scanner.get();
     }
     return scanners;
 }
 
-void InputArtifactScanner::scanForScannerFileDependencies(DependencyScanner *scanner,
-        Artifact *inputArtifact, FileResourceBase *fileToBeScanned,
-        QList<FileResourceBase *> *filesToScan,
-        InputArtifactScannerContext::ScannerResolvedDependenciesCache &cache)
+void InputArtifactScanner::scanForScannerFileDependencies(
+    DependencyScanner *scanner,
+    Artifact *inputArtifact,
+    FileResourceBase *fileToBeScanned,
+    QList<FileResourceBase *> *filesToScan,
+    InputArtifactScannerContext::ScannerKeyCacheItem &cache)
 {
     qCDebug(lcDepScan) << "file" << fileToBeScanned->filePath();
 
-    const bool cacheHit = cache.valid;
+    const bool cacheHit = !!cache;
     if (!cacheHit) {
-        cache.valid = true;
-        cache.searchPaths = scanner->collectSearchPaths(inputArtifact);
+        cache.emplace();
+        cache->searchPaths = scanner->collectSearchPaths(inputArtifact);
     }
     qCDebug(lcDepScan) << "include paths (cache" << (cacheHit ? "hit)" : "miss)");
-    for (const QString &s : std::as_const(cache.searchPaths))
+    for (const QString &s : std::as_const(cache->searchPaths))
         qCDebug(lcDepScan) << "    " << s;
 
     const QString &filePathToBeScanned = fileToBeScanned->filePath();
     RawScanResults::ScanData &scanData = m_rawScanResults.findScanData(fileToBeScanned, scanner,
                                                                        m_artifact->properties);
     if (scanData.lastScanTime < fileToBeScanned->timestamp()) {
-        try {
-            qCDebug(lcDepScan) << "scanning" << FileInfo::fileName(filePathToBeScanned);
-            scanWithScannerPlugin(scanner, inputArtifact, fileToBeScanned, &scanData.rawScanResult);
-            scanData.lastScanTime = FileTime::currentTime();
-        } catch (const ErrorInfo &error) {
-            m_logger.printWarning(error);
-            return;
-        }
+        qCDebug(lcDepScan) << "scanning" << FileInfo::fileName(filePathToBeScanned);
+        scanWithScannerPlugin(scanner, inputArtifact, fileToBeScanned, &scanData.rawScanResult);
+        scanData.lastScanTime = FileTime::currentTime();
     }
 
-    resolveScanResultDependencies(inputArtifact, scanData.rawScanResult, filesToScan, cache);
+    resolveScanResultDependencies(inputArtifact, scanData.rawScanResult, filesToScan, *cache);
 }
 
-void InputArtifactScanner::resolveScanResultDependencies(const Artifact *inputArtifact,
-        const RawScanResult &scanResult, QList<FileResourceBase *> *artifactsToScan,
-        InputArtifactScannerContext::ScannerResolvedDependenciesCache &cache)
+void InputArtifactScanner::resolveScanResultDependencies(
+    const Artifact *inputArtifact,
+    const RawScanResult &scanResult,
+    QList<FileResourceBase *> *artifactsToScan,
+    InputArtifactScannerContext::ScannerKeyCacheData &cache)
 {
     auto getResolvedDependency = [inputArtifact, &cache](const RawScannedDependency &dependency)
             -> ResolvedDependency*
@@ -273,13 +283,13 @@ void InputArtifactScanner::resolveScanResultDependencies(const Artifact *inputAr
         const QString &dependencyFilePath = dependency.filePath();
         InputArtifactScannerContext::ResolvedDependencyCacheItem &cachedResolvedDependencyItem
                 = cache.resolvedDependenciesCache[dependency.dirPath()][dependency.fileName()];
-        ResolvedDependency &resolvedDependency = cachedResolvedDependencyItem.resolvedDependency;
-        if (cachedResolvedDependencyItem.valid) {
+        if (cachedResolvedDependencyItem) {
+            ResolvedDependency &resolvedDependency = *cachedResolvedDependencyItem;
             if (resolvedDependency.filePath.isEmpty())
                 return nullptr;
             return &resolvedDependency;
         }
-        cachedResolvedDependencyItem.valid = true;
+        ResolvedDependency &resolvedDependency = cachedResolvedDependencyItem.emplace();
 
         if (FileInfo::isAbsolute(dependencyFilePath)) {
             resolveDepencency(dependency, inputArtifact->product.get(), &resolvedDependency);
@@ -396,11 +406,5 @@ void InputArtifactScanner::scanWithScannerPlugin(DependencyScanner *scanner,
         scanResult->deps.emplace_back(s);
 }
 
-InputArtifactScannerContext::DependencyScannerCacheItem::DependencyScannerCacheItem() : valid(false)
-{
-}
-
-InputArtifactScannerContext::DependencyScannerCacheItem::~DependencyScannerCacheItem() = default;
-
 } // namespace Internal
 } // namespace qbs
