@@ -38,20 +38,42 @@
 ****************************************************************************/
 
 #include "qttools.h"
+#include "porting.h"
 
 #include <QtCore/qprocess.h>
 
+namespace std {
+
+size_t hash<QVariant>::operator()(const QVariant &v) const noexcept
+{
+    switch (v.userType()) {
+    case QMetaType::UnknownType: return 0;
+    case QMetaType::Bool: return std::hash<bool>()(v.toBool());
+    case QMetaType::Int: return std::hash<int>()(v.toInt());
+    case QMetaType::UInt: return std::hash<uint>()(v.toUInt());
+    case QMetaType::QString: return std::hash<QString>()(v.toString());
+    case QMetaType::QStringList: return std::hash<QStringList>()(v.toStringList());
+    case QMetaType::QVariantList: return std::hash<QVariantList>()(v.toList());
+    case QMetaType::QVariantMap: return std::hash<QVariantMap>()(v.toMap());
+    case QMetaType::QVariantHash: return std::hash<QVariantHash>()(v.toHash());
+    default:
+        QBS_ASSERT("Unsupported variant type" && false, return 0);
+    }
+}
+
+} // namespace std
+
 QT_BEGIN_NAMESPACE
 
-uint qHash(const QStringList &list)
+qbs::QHashValueType qHash(const QStringList &list)
 {
-    uint s = 0;
+    qbs::QHashValueType s = 0;
     for (const QString &n : list)
         s ^= qHash(n) + 0x9e3779b9 + (s << 6) + (s >> 2);
     return s;
 }
 
-uint qHash(const QProcessEnvironment &env)
+qbs::QHashValueType qHash(const QProcessEnvironment &env)
 {
     return qHash(env.toStringList());
 }
