@@ -34,11 +34,13 @@
 #include <tools/qttools.h>
 
 #include <QtCore/qdiriterator.h>
+#include <QtCore/qjsonarray.h>
 #include <QtCore/qjsondocument.h>
 #include <QtCore/qjsonobject.h>
 #include <QtXml/qdom.h>
 
 #include <regex>
+#include <unordered_set>
 
 #define WAIT_FOR_NEW_TIMESTAMP() waitForNewTimestamp(testDataDir)
 
@@ -233,7 +235,7 @@ void TestBlackboxApple::appleMultiConfig()
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiapp.app/Contents/PkgInfo").isRegularFile());
 
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiappmultivariant.app/Contents/MacOS/"
-                                                "fatmultiappmultivariant").isFileSymLink());
+                                                "fatmultiappmultivariant").isRegularFile());
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiappmultivariant.app/Contents/MacOS/"
                                                 "fatmultiappmultivariant_debug").isExecutable());
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiappmultivariant.app/Contents/MacOS/"
@@ -255,7 +257,8 @@ void TestBlackboxApple::appleMultiConfig()
         QVERIFY(QFileInfo2(defaultInstallRoot + "/multilib.framework/Versions/Current").isDirSymLink());
 
         for (const QString variant : { "release", "debug", "profiling" }) {
-            for (const QString arch : { "x86_64" }) {
+            const QString arch = HostOsInfo::hostOSArchitecture() == "arm64" ? "arm64" : "x86_64";
+            for (const QString &arch : {arch}) {
                 QProcess process;
                 process.setProgram("/usr/bin/arch");
                 process.setArguments({
@@ -292,7 +295,7 @@ void TestBlackboxApple::appleMultiConfig()
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiapp.app/PkgInfo").isRegularFile());
 
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiappmultivariant.app/"
-                                                "fatmultiappmultivariant").isFileSymLink());
+                                                "fatmultiappmultivariant").isRegularFile());
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiappmultivariant.app/"
                                                 "fatmultiappmultivariant_debug").isExecutable());
         QVERIFY(QFileInfo2(defaultInstallRoot + "/fatmultiappmultivariant.app/"
@@ -493,194 +496,88 @@ void TestBlackboxApple::bundleStructure()
     rmDirR(relativeBuildDir());
     const int status = runQbs(params);
     if (status != 0) {
-        QVERIFY2(m_qbsStderr.contains("Bundle product type "
-                                      + productTypeIdentifier.toLatin1()
-                                      + " is not supported."),
-                 m_qbsStderr.constData());
+        QVERIFY2(
+            m_qbsStderr.contains(
+                "Bundle product type " + productTypeIdentifier.toLatin1() + " is not supported."),
+            m_qbsStderr.constData());
         return;
     }
 
     QCOMPARE(status, 0);
 
-    if (m_qbsStdout.contains("bundle.isShallow: false")) {
-        // Test shallow bundles detection - bundles are not shallow only on macOS, so also check
-        // the qbs.targetOS property
-        QVERIFY2(m_qbsStdout.contains("qbs.targetOS: macos"), m_qbsStdout);
-        if (productName == "A") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents/MacOS").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents/MacOS/A").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents/PkgInfo").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents/Resources").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Contents/Resources/resource.txt").isRegularFile());
-        }
+    const bool enableCodeSigning = m_qbsStdout.contains("enableCodeSigning: true");
+    const bool isShallow = m_qbsStdout.contains("bundle.isShallow: true");
 
-        if (productName == "B") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/B").isFileSymLink());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Headers").isDirSymLink());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Modules").isDirSymLink());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/B.framework/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/PrivateHeaders").isDirSymLink());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Resources").isDirSymLink());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/B").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/Headers/dummy.h").isRegularFile());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/Modules").isRegularDir());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/Modules/module.modulemap").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/Resources").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/A/Resources/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Versions/Current").isDirSymLink());
-        }
-
-        if (productName == "C") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/C").isFileSymLink());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Headers").isDirSymLink());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Modules").isDirSymLink());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/C.framework/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/PrivateHeaders").isDirSymLink());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Resources").isDirSymLink());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/C").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/Headers/dummy.h").isRegularFile());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/Modules").isRegularDir());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/Modules/module.modulemap").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/Resources").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/A/Resources/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Versions/Current").isDirSymLink());
-        }
-
-        if (productName == "D") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Contents").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Contents/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Contents/MacOS").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Contents/MacOS/D").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/D.bundle/Contents/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Contents/Resources").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Contents/Resources/resource.txt").isRegularFile());
-        }
-
-        if (productName == "E") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Contents").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Contents/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Contents/MacOS").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Contents/MacOS/E").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/E.appex/Contents/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Contents/Resources").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Contents/Resources/resource.txt").isRegularFile());
-        }
-
-        if (productName == "F") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Contents").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Contents/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Contents/MacOS").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Contents/MacOS/F").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/F.xpc/Contents/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Contents/Resources").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Contents/Resources/resource.txt").isRegularFile());
-        }
-
-        if (productName == "G") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/G").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/G/ContentInfo.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/G/Contents/resource.txt").isRegularFile());
-        }
-    } else if (m_qbsStdout.contains("bundle.isShallow: true")) {
-        QVERIFY2(m_qbsStdout.contains("qbs.targetOS:"), m_qbsStdout);
-        QVERIFY2(!m_qbsStdout.contains("qbs.targetOS: macos"), m_qbsStdout);
-        if (productName == "A") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/A").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/Info.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/PkgInfo").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/A.app/resource.txt").isRegularFile());
-        }
-
-        if (productName == "B") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/B").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Headers/dummy.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Info.plist").isRegularFile());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Modules").isRegularDir());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/Modules/module.modulemap").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/B.framework/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/B.framework/resource.txt").isRegularFile());
-        }
-
-        if (productName == "C") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/C").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Headers/dummy.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Info.plist").isRegularFile());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Modules").isRegularDir());
-            //QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/Modules/module.modulemap").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/C.framework/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/C.framework/resource.txt").isRegularFile());
-        }
-
-        if (productName == "D") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/D").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Headers/dummy.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/Info.plist").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/D.bundle/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/D.bundle/resource.txt").isRegularFile());
-        }
-
-        if (productName == "E") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/E").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Headers/dummy.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/Info.plist").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/E.appex/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/E.appex/resource.txt").isRegularFile());
-        }
-
-        if (productName == "F") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/F").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Headers").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Headers/dummy.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/Info.plist").isRegularFile());
-            QVERIFY(!QFileInfo2(defaultInstallRoot + "/F.xpc/PkgInfo").exists());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/PrivateHeaders").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/PrivateHeaders/dummy_p.h").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/F.xpc/resource.txt").isRegularFile());
-        }
-
-        if (productName == "G") {
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/G").isRegularDir());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/G/ContentInfo.plist").isRegularFile());
-            QVERIFY(QFileInfo2(defaultInstallRoot + "/G/Contents/resource.txt").isRegularFile());
-        }
+    QString expectedStructureFile;
+    if (!isShallow) {
+        QVERIFY2(enableCodeSigning, m_qbsStdout);
+        expectedStructureFile = "deep.json";
     } else {
-        QVERIFY2(false, qPrintable(m_qbsStdout));
+        expectedStructureFile = enableCodeSigning ? "shallow-signed.json" : "shallow-unsigned.json";
+    }
+
+    const QString expectedStructurePath = testDataDir + "/bundle-structure/"
+                                          + expectedStructureFile;
+    QFile structureFile(expectedStructurePath);
+    QVERIFY2(
+        structureFile.open(QIODevice::ReadOnly),
+        qPrintable(QString("Could not open %1").arg(expectedStructurePath)));
+
+    QJsonParseError parseError;
+    const auto expectedStructure = QJsonDocument::fromJson(structureFile.readAll(), &parseError);
+    QVERIFY2(parseError.error == QJsonParseError::NoError, qPrintable(parseError.errorString()));
+
+    const auto allProducts = expectedStructure.object();
+    QVERIFY2(
+        allProducts.contains(productName),
+        qPrintable(QString("Product %1 not found in %2").arg(productName, expectedStructureFile)));
+
+    const auto productStructure = allProducts.value(productName).toObject();
+    const QString bundleExtension = productStructure.value("extension").toString();
+    const QString bundlePath = defaultInstallRoot + "/" + productName + bundleExtension;
+
+    std::unordered_set<QString> expectedPaths;
+
+    // Check directories
+    const auto directories = productStructure.value("directories").toArray();
+    for (const auto &dir : directories) {
+        const QString dirPath = bundlePath + "/" + dir.toString();
+        expectedPaths.insert(dirPath);
+        QVERIFY2(
+            QFileInfo2(dirPath).isRegularDir(),
+            qPrintable(QString("Directory does not exist: %1").arg(dirPath)));
+    }
+
+    // Check files
+    const auto files = productStructure.value("files").toArray();
+    for (const auto &file : files) {
+        const QString filePath = bundlePath + "/" + file.toString();
+        expectedPaths.insert(filePath);
+        QVERIFY2(
+            QFileInfo2(filePath).isRegularFile(),
+            qPrintable(QString("File does not exist: %1").arg(filePath)));
+    }
+
+    // Check symlinks
+    const auto symlinks = productStructure.value("symlinks").toArray();
+    for (const auto &symlink : symlinks) {
+        const QString symlinkPath = bundlePath + "/" + symlink.toString();
+        expectedPaths.insert(symlinkPath);
+        QVERIFY2(
+            QFileInfo2(symlinkPath).isSymLink(),
+            qPrintable(QString("Symlink does not exist: %1").arg(symlinkPath)));
+    }
+
+    // Check that there are no extra entries beyond what's expected
+    QDirIterator it(
+        bundlePath, QDir::AllEntries | QDir::NoDotAndDotDot, QDirIterator::Subdirectories);
+    while (it.hasNext()) {
+        const QString absolutePath = it.next();
+        const QString relativePath = QDir(bundlePath).relativeFilePath(absolutePath);
+
+        QVERIFY2(
+            expectedPaths.count(absolutePath) != 0,
+            qPrintable(QString("Unexpected file/directory found in bundle: %1").arg(relativePath)));
     }
 }
 
@@ -688,7 +585,6 @@ void TestBlackboxApple::bundleStructure_data()
 {
     QTest::addColumn<QString>("productName");
     QTest::addColumn<QString>("productTypeIdentifier");
-    QTest::addColumn<bool>("isShallow");
 
     QTest::newRow("A") << "A" << "com.apple.product-type.application";
     QTest::newRow("ABadApple") << "ABadApple" << "com.apple.product-type.will.never.exist.ever.guaranteed";
@@ -701,6 +597,111 @@ void TestBlackboxApple::bundleStructure_data()
     QTest::newRow("G") << "G" << "com.apple.product-type.in-app-purchase-content";
 }
 
+void TestBlackboxApple::mainBundle()
+{
+    QFETCH(bool, multiplexArchitectures);
+    QFETCH(bool, multiplexBuildVariants);
+
+    QDir::setCurrent(testDataDir + "/main-bundle");
+    QbsRunParameters params(QStringList{"qbs.installPrefix:''"});
+
+    const auto xcodeVersion = findXcodeVersion();
+    if (!xcodeVersion)
+        QSKIP("requires Xcode profile");
+
+    params.arguments << "project.xcodeVersion:" + xcodeVersion->toString();
+    if (multiplexArchitectures)
+        params.arguments << "project.multiplexArchitectures:true";
+    if (multiplexBuildVariants)
+        params.arguments << "project.multiplexBuildVariants:true";
+
+    rmDirR(relativeBuildDir());
+    const int status = runQbs(params);
+    QCOMPARE(status, 0);
+
+    const bool enableCodeSigning = m_qbsStdout.contains("enableCodeSigning: true");
+    const bool isShallow = m_qbsStdout.contains("bundle.isShallow: true");
+
+    QString expectedStructureFile;
+    if (!isShallow) {
+        QVERIFY2(enableCodeSigning, m_qbsStdout);
+        expectedStructureFile = "deep.json";
+    } else {
+        expectedStructureFile = enableCodeSigning ? "shallow-signed.json" : "shallow-unsigned.json";
+    }
+
+    const QString expectedStructurePath = testDataDir + "/main-bundle/" + expectedStructureFile;
+    QFile structureFile(expectedStructurePath);
+    QVERIFY2(
+        structureFile.open(QIODevice::ReadOnly),
+        qPrintable(QString("Could not open %1").arg(expectedStructurePath)));
+
+    QJsonParseError parseError;
+    const auto expectedStructure = QJsonDocument::fromJson(structureFile.readAll(), &parseError);
+    QVERIFY2(parseError.error == QJsonParseError::NoError, qPrintable(parseError.errorString()));
+
+    const auto structure = expectedStructure.object();
+    const QString appPath = defaultInstallRoot + "/MainApp.app";
+
+    std::unordered_set<QString> expectedPaths;
+
+    // Check directories
+    const auto directories = structure.value("directories").toArray();
+    for (const auto &dir : directories) {
+        const QString dirPath = appPath + "/" + dir.toString();
+        expectedPaths.insert(dirPath);
+        QVERIFY2(
+            QFileInfo2(dirPath).isRegularDir(),
+            qPrintable(QString("Directory does not exist: %1").arg(dirPath)));
+    }
+
+    // Check files
+    const auto files = structure.value("files").toArray();
+    for (const auto &file : files) {
+        const QString filePath = appPath + "/" + file.toString();
+        expectedPaths.insert(filePath);
+        QVERIFY2(
+            QFileInfo2(filePath).isRegularFile(),
+            qPrintable(QString("File does not exist: %1").arg(filePath)));
+    }
+
+    // Check symlinks
+    const auto symlinks = structure.value("symlinks").toArray();
+    for (const auto &symlink : symlinks) {
+        const QString symlinkPath = appPath + "/" + symlink.toString();
+        expectedPaths.insert(symlinkPath);
+        QVERIFY2(
+            QFileInfo2(symlinkPath).isSymLink(),
+            qPrintable(QString("Symlink does not exist: %1").arg(symlinkPath)));
+    }
+
+    // Check that there are no extra entries beyond what's expected
+    QDirIterator it(appPath, QDir::AllEntries | QDir::NoDotAndDotDot, QDirIterator::Subdirectories);
+    while (it.hasNext()) {
+        const QString absolutePath = it.next();
+        const QString relativePath = QDir(appPath).relativeFilePath(absolutePath);
+
+        // skip debug build variants manually
+        if (multiplexBuildVariants && QFileInfo(absolutePath).baseName().endsWith("_debug"))
+            continue;
+
+        QVERIFY2(
+            expectedPaths.count(absolutePath) != 0,
+            qPrintable(QString("Unexpected file/directory found in bundle: %1").arg(relativePath)));
+    }
+}
+
+void TestBlackboxApple::mainBundle_data()
+{
+    QTest::addColumn<bool>("multiplexArchitectures");
+    QTest::addColumn<bool>("multiplexBuildVariants");
+
+    QTest::newRow("no multiplexing") << false << false;
+    QTest::newRow("multiplex architectures") << true << false;
+    QTest::newRow("multiplex build variants") << false << true;
+    QTest::newRow("multiplex architectures and build variants") << true << true;
+}
+
 void TestBlackboxApple::byteArrayInfoPlist()
 {
     QDir::setCurrent(testDataDir + "/byteArrayInfoPlist");
@@ -726,15 +727,18 @@ void TestBlackboxApple::codesign()
     QFETCH(bool, multiArch);
     QFETCH(bool, multiVariant);
 
-    const auto xcodeVersion = findXcodeVersion();
+    const auto xcode = findXcode();
 
-    if (!xcodeVersion)
+    if (!xcode)
         QSKIP("requires Xcode profile");
 
+    const auto xcodeVersion = qbs::Version::fromString(xcode->value("version").toString());
+    const bool isSigned = xcode->value("isSignedByDefault").toBool() || enableSigning;
+
     QDir::setCurrent(testDataDir + "/codesign");
     QbsRunParameters params(QStringList{"qbs.installPrefix:''"});
     // the test can't use xcode module to determine version itself
-    params.arguments << QStringLiteral("project.xcodeVersion:") + xcodeVersion->toString();
+    params.arguments << QStringLiteral("project.xcodeVersion:") + xcodeVersion.toString();
     params.arguments << QStringLiteral("project.isBundle:%1").arg(isBundle ? "true" : "false");
     params.arguments << QStringLiteral("project.enableSigning:%1")
                             .arg(enableSigning ? "true" : "false");
@@ -745,8 +749,7 @@ void TestBlackboxApple::codesign()
     rmDirR(relativeBuildDir());
     QCOMPARE(runQbs(params), 0);
 
-    const int codeSignCount =
-            QString::fromUtf8(m_qbsStdout).count(QStringLiteral("codesign"));
+    const int codeSignCount = QString::fromUtf8(m_qbsStdout).count(QStringLiteral("codesign "));
     QCOMPARE(codeSignCount, expectedCount);
 
     const auto appName = isBundle ? QStringLiteral("A.app") : QStringLiteral("A");
@@ -754,8 +757,8 @@ void TestBlackboxApple::codesign()
     QVERIFY(QFileInfo(appPath).exists());
     auto codeSignInfo = getCodeSignInfo(appPath);
     QVERIFY(codeSignInfo.first != CodeSignResult::Failed);
-    QCOMPARE(codeSignInfo.first == CodeSignResult::Signed, enableSigning);
-    QCOMPARE(codeSignInfo.second.isEmpty(), !enableSigning);
+    QCOMPARE(codeSignInfo.first == CodeSignResult::Signed, isSigned);
+    QCOMPARE(codeSignInfo.second.isEmpty(), !isSigned);
     if (!codeSignInfo.second.isEmpty()) {
         QVERIFY(codeSignInfo.second.contains(QByteArrayLiteral("Executable")));
         QVERIFY(codeSignInfo.second.contains(QByteArrayLiteral("Identifier")));
@@ -768,8 +771,8 @@ void TestBlackboxApple::codesign()
     QVERIFY(QFileInfo(libPath).exists());
     codeSignInfo = getCodeSignInfo(libPath);
     QVERIFY(codeSignInfo.first != CodeSignResult::Failed);
-    QCOMPARE(codeSignInfo.first == CodeSignResult::Signed, enableSigning);
-    QCOMPARE(codeSignInfo.second.isEmpty(), !enableSigning);
+    QCOMPARE(codeSignInfo.first == CodeSignResult::Signed, isSigned);
+    QCOMPARE(codeSignInfo.second.isEmpty(), !isSigned);
     if (!codeSignInfo.second.isEmpty()) {
         QVERIFY(codeSignInfo.second.contains(QByteArrayLiteral("Executable")));
         QVERIFY(codeSignInfo.second.contains(QByteArrayLiteral("Identifier")));
@@ -781,6 +784,20 @@ void TestBlackboxApple::codesign()
     QVERIFY(QFileInfo(pluginPath).isDir() == isBundle);
     codeSignInfo = getCodeSignInfo(pluginPath);
     QVERIFY(codeSignInfo.first != CodeSignResult::Failed);
+    QCOMPARE(codeSignInfo.first == CodeSignResult::Signed, isSigned);
+    QCOMPARE(codeSignInfo.second.isEmpty(), !isSigned);
+    if (!codeSignInfo.second.isEmpty()) {
+        QVERIFY(codeSignInfo.second.contains(QByteArrayLiteral("Executable")));
+        QVERIFY(codeSignInfo.second.contains(QByteArrayLiteral("Identifier")));
+        QCOMPARE(codeSignInfo.second.value(QByteArrayLiteral("Signature")), "adhoc");
+    }
+
+    const auto staticLibName = isBundle ? QStringLiteral("D.framework") : QStringLiteral("libD.a");
+    const auto staticLibPath = defaultInstallRoot + "/" + staticLibName;
+    QVERIFY(QFileInfo(staticLibPath).exists());
+    codeSignInfo = getCodeSignInfo(staticLibPath);
+    QVERIFY(codeSignInfo.first != CodeSignResult::Failed);
+    // static libraries are not signed by default so we check for enableSigning instead of isSigned
     QCOMPARE(codeSignInfo.first == CodeSignResult::Signed, enableSigning);
     QCOMPARE(codeSignInfo.second.isEmpty(), !enableSigning);
     if (!codeSignInfo.second.isEmpty()) {
@@ -800,17 +817,24 @@ void TestBlackboxApple::codesign_data()
 
     QTest::newRow("standalone, unsigned") << 0 << false << false << false << false;
     QTest::newRow("bundle, unsigned") << 0 << true << false << false << false;
-    QTest::newRow("standalone, signed") << 3 << false << true << false << false;
-    QTest::newRow("bundle, signed") << 3 << true << true << false << false;
+    QTest::newRow("standalone, signed") << 4 << false << true << false << false;
+    QTest::newRow("bundle, signed") << 4 << true << true << false << false;
     // here we only sign the resulting lipo artifact
-    QTest::newRow("standalone, signed, multiarch") << 3 << false << true << true << false;
-    QTest::newRow("bundle, signed, multiarch") << 3 << true << true << true << false;
+    QTest::newRow("standalone, signed, multiarch") << 4 << false << true << true << false;
+    QTest::newRow("bundle, signed, multiarch") << 4 << true << true << true << false;
     // here we sign all artifacts
-    QTest::newRow("standalone, signed, multivariant") << 15 << false << true << false << true;
-    QTest::newRow("bundle, signed, multivariant") << 15 << true << true << false << true;
+    QTest::newRow("standalone, signed, multivariant") << 8 << false << true << false << true;
+    QTest::newRow("bundle, signed, multivariant") << 8 << true << true << false << true;
     QTest::newRow("standalone, signed, multiarch, multivariant")
-        << 15 << false << true << true << true;
-    QTest::newRow("bundle, signed, multiarch, multivariant") << 15 << true << true << true << true;
+        << 8 << false << true << true << true;
+    QTest::newRow("bundle, signed, multiarch, multivariant") << 8 << true << true << true << true;
+}
+
+void TestBlackboxApple::codesignDestinationDirectory()
+{
+    // checks that codesign does not break when destinationDirectory is set
+    QDir::setCurrent(testDataDir + "/codesign-destination-directory");
+    QCOMPARE(runQbs(), 0);
 }
 
 void TestBlackboxApple::deploymentTarget()
@@ -872,54 +896,65 @@ void TestBlackboxApple::deploymentTarget_data()
     QTest::addColumn<QString>("lflags");
 
     const auto xcodeVersion = findXcodeVersion();
+    QTest::newRow("macos arm64") << "macosx" << macos << "arm64"
+                                 << "-triple arm64-apple-macosx11.0"
+                                 << "-platform_version macos 11.0.0";
     if (xcodeVersion < qbs::Version(10)) {
         QTest::newRow("macos x86") << "macosx" << macos << "x86"
                                    << "-triple i386-apple-macosx10.6"
                                    << "-macosx_version_min 10.6";
     }
     QTest::newRow("macos x86_64") << "macosx" << macos << "x86_64"
-                         << "-triple x86_64-apple-macosx10.6"
-                         << "10.6";
+                                  << "-triple x86_64-apple-macosx10.6"
+                                  << "10.6";
 
     if (xcodeVersion >= qbs::Version(6))
         QTest::newRow("macos x86_64h") << "macosx" << macos << "x86_64h"
-                             << "-triple x86_64h-apple-macosx10.12"
-                             << "10.12";
+                                       << "-triple x86_64h-apple-macosx10.12"
+                                       << "10.12";
 
-    QTest::newRow("ios armv7a") << "iphoneos" << ios << "armv7a"
-                         << "-triple thumbv7-apple-ios6.0"
-                         << "6.0";
-    QTest::newRow("ios armv7s") << "iphoneos" <<ios << "armv7s"
-                         << "-triple thumbv7s-apple-ios7.0"
-                         << "7.0";
+    if (xcodeVersion < qbs::Version(16)) {
+        QTest::newRow("ios armv7a") << "iphoneos" << ios << "armv7a"
+                                    << "-triple thumbv7-apple-ios6.0"
+                                    << "6.0";
+        QTest::newRow("ios armv7s") << "iphoneos" << ios << "armv7s"
+                                    << "-triple thumbv7s-apple-ios7.0"
+                                    << "7.0";
+    }
     if (xcodeVersion >= qbs::Version(5))
-        QTest::newRow("ios arm64") << "iphoneos" <<ios << "arm64"
-                             << "-triple arm64-apple-ios7.0"
-                             << "7.0";
-    QTest::newRow("ios-simulator x86") << "iphonesimulator" << ios_sim << "x86"
-                             << "-triple i386-apple-ios6.0"
-                             << "6.0";
+        QTest::newRow("ios arm64") << "iphoneos" << ios << "arm64"
+                                   << "-triple arm64-apple-ios7.0"
+                                   << "7.0";
+    if (xcodeVersion < qbs::Version(16)) {
+        QTest::newRow("ios-simulator x86") << "iphonesimulator" << ios_sim << "x86"
+                                           << "-triple i386-apple-ios6.0"
+                                           << "6.0";
+        if (xcodeVersion >= qbs::Version(5))
+            QTest::newRow("ios-simulator x86_64") << "iphonesimulator" << ios_sim << "x86_64"
+                                                  << "-triple x86_64-apple-ios7.0"
+                                                  << "7.0";
+    }
     if (xcodeVersion >= qbs::Version(5))
-        QTest::newRow("ios-simulator x86_64") << "iphonesimulator" << ios_sim << "x86_64"
-                                 << "-triple x86_64-apple-ios7.0"
-                                 << "7.0";
+        QTest::newRow("ios-simulator arm64") << "iphonesimulator" << ios_sim << "arm64"
+                                             << "-triple arm64-apple-ios7.0"
+                                             << "7.0";
 
     if (xcodeVersion >= qbs::Version(7)) {
         if (xcodeVersion >= qbs::Version(7, 1)) {
             QTest::newRow("tvos arm64") << "appletvos" << tvos << "arm64"
-                                 << "-triple arm64-apple-tvos9.0"
-                                 << "9.0";
+                                        << "-triple arm64-apple-tvos9.0"
+                                        << "9.0";
             QTest::newRow("tvos-simulator x86_64") << "appletvsimulator" << tvos_sim << "x86_64"
-                                     << "-triple x86_64-apple-tvos9.0"
-                                     << "9.0";
+                                                   << "-triple x86_64-apple-tvos9.0"
+                                                   << "9.0";
         }
 
         QTest::newRow("watchos armv7k") << "watchos" << watchos << "armv7k"
-                             << "-triple thumbv7k-apple-watchos2.0"
-                             << "2.0";
+                                        << "-triple thumbv7k-apple-watchos2.0"
+                                        << "2.0";
         QTest::newRow("watchos-simulator x86") << "watchsimulator" << watchos_sim << "x86"
-                                 << "-triple i386-apple-watchos2.0"
-                                 << "2.0";
+                                               << "-triple i386-apple-watchos2.0"
+                                               << "2.0";
     }
 }
 
