@@ -44,12 +44,17 @@
 
 namespace qbsBenchmarker {
 
-ValgrindRunner::ValgrindRunner(Activities activities, QString testProject,
-                               const QString &qbsBuildDir, const QString &baseOutputDir)
+ValgrindRunner::ValgrindRunner(
+    Activities activities,
+    QString testProject,
+    const QString &qbsBuildDir,
+    const QString &baseOutputDir,
+    bool sequential)
     : m_activities(activities)
     , m_testProject(std::move(testProject))
-    , m_qbsBinary(qbsBuildDir + "/bin/qbs")
+    , m_qbsBinary(qbsBuildDir + "/benchmarker/install-root/bin/qbs")
     , m_baseOutputDir(baseOutputDir)
+    , m_sequential(sequential)
 {
     if (!QDir::root().mkpath(m_baseOutputDir))
         throw Exception(QStringLiteral("Failed to create directory '%1'.").arg(baseOutputDir));
@@ -57,13 +62,23 @@ ValgrindRunner::ValgrindRunner(Activities activities, QString testProject,
 
 void ValgrindRunner::run()
 {
+    if (m_sequential) {
+        if (m_activities & ActivityResolving)
+            traceResolving();
+        if (m_activities & ActivityRuleExecution)
+            traceRuleExecution();
+        if (m_activities & ActivityNullBuild)
+            traceNullBuild();
+        return;
+    }
+
     std::deque<QFuture<void>> futures;
     if (m_activities & ActivityResolving)
-        futures.push_back(QtConcurrent::run(this, &ValgrindRunner::traceResolving));
+        futures.push_back(QtConcurrent::run([this]{ traceResolving(); }));
     if (m_activities & ActivityRuleExecution)
-        futures.push_back(QtConcurrent::run(this, &ValgrindRunner::traceRuleExecution));
+        futures.push_back(QtConcurrent::run([this]{ traceRuleExecution(); }));
     if (m_activities & ActivityNullBuild)
-        futures.push_back(QtConcurrent::run(this, &ValgrindRunner::traceNullBuild));
+        futures.push_back(QtConcurrent::run([this]{ traceNullBuild(); }));
     while (!futures.empty()) {
         futures.front().waitForFinished();
         futures.pop_front();
@@ -100,12 +115,11 @@ void ValgrindRunner::traceActivity(Activity activity, const QString &buildDirCal
 {
     QString activityString;
     QString qbsCommand;
-    bool dryRun;
+    bool dryRun = false;
     switch (activity) {
     case ActivityResolving:
         activityString = "resolving";
         qbsCommand = "resolve";
-        dryRun = false;
         break;
     case ActivityRuleExecution:
         activityString = "rule-execution";
@@ -115,19 +129,31 @@ void ValgrindRunner::traceActivity(Activity activity, const QString &buildDirCal
     case ActivityNullBuild:
         activityString = "null-build";
         qbsCommand = "build";
-        dryRun = false;
         break;
     }
 
     const QString outFileCallgrind = m_baseOutputDir + "/outfile." + activityString + ".callgrind";
     const QString outFileMassif = m_baseOutputDir + "/outfile." + activityString + ".massif";
-    QFuture<qint64> callGrindFuture = QtConcurrent::run(this, &ValgrindRunner::runCallgrind,
-            qbsCommand, buildDirCallgrind, dryRun, outFileCallgrind);
-    QFuture<qint64> massifFuture = QtConcurrent::run(this, &ValgrindRunner::runMassif, qbsCommand,
-            buildDirMassif, dryRun, outFileMassif);
-    callGrindFuture.waitForFinished();
-    massifFuture.waitForFinished();
-    addToResults(ValgrindResult(activity, callGrindFuture.result(), massifFuture.result()));
+    qint64 instructionCount = 0;
+    qint64 memUsage = 0;
+    const auto callgrindRunner = [this, qbsCommand, buildDirCallgrind, dryRun, outFileCallgrind] {
+        return runCallgrind(qbsCommand, buildDirCallgrind, dryRun, outFileCallgrind);
+    };
+    const auto massifRunner = [this, qbsCommand, buildDirMassif, dryRun, outFileMassif] {
+        return runMassif(qbsCommand, buildDirMassif, dryRun, outFileMassif);
+    };
+    if (m_sequential) {
+        instructionCount = callgrindRunner();
+        memUsage = massifRunner();
+    } else {
+        QFuture<qint64> callGrindFuture = QtConcurrent::run(callgrindRunner);
+        QFuture<qint64> massifFuture = QtConcurrent::run(massifRunner);
+        callGrindFuture.waitForFinished();
+        instructionCount = callGrindFuture.result();
+        massifFuture.waitForFinished();
+        memUsage = massifFuture.result();
+    }
+    addToResults(ValgrindResult(activity, instructionCount, memUsage));
 }
 
 QStringList ValgrindRunner::qbsCommandLine(const QString &command, const QString &buildDir,
@@ -221,7 +247,7 @@ qint64 ValgrindRunner::runMassif(const QString &qbsCommand, const QString &build
         throw Exception(exceptionStringPattern.arg("No peak marker found"));
     while (!buffer.atEnd()) {
         const QByteArray line = buffer.readLine().simplified();
-        if (!line.startsWith(peakSnapshot + ' '))
+        if (!line.startsWith(QByteArray(peakSnapshot + ' ')))
             continue;
         const QList<QByteArray> entries = line.split(' ');
         if (entries.size() != 6) {
